//---------------------------------Spheral++----------------------------------//
// GaussLegendreValues
//
// Positive ordinates for the Gauss-Legendre quadrature, accurate to 20 digits
//----------------------------------------------------------------------------//
#include "GaussLegendreValues.hh"

#include "Utilities/DBC.hh"

namespace Spheral {

int
GaussLegendreValues::
numOrdinatesForOrder(const int order) {
  // (order + 1) / 2, with an extra 1/2 for ceil operation once truncated
  const auto minOrdinates = (int)(0.5 * (static_cast<double>(order) + 2.));
  for (auto i = 1; i < 1025; i *= 2) {
    if (i >= minOrdinates) return i;
  }
  VERIFY2(false, "order is too high for number of ordinates supported");
  return -1;
}

int
GaussLegendreValues::
orderForNumOrdinates(const int numOrdinates) {
  return 2 * numOrdinates - 1;
}

const std::vector<double>&
GaussLegendreValues::
getValues(const int numOrdinates) {
    switch (numOrdinates) {
    case 2: return values2;
    case 4: return values4;
    case 8: return values8;
    case 16: return values16;
    case 32: return values32;
    case 64: return values64;
    case 128: return values128;
    case 256: return values256;
    case 512: return values512;
    case 1024: return values1024;
    default: VERIFY2(false, "GaussLegendreValues: quadrature not found");
    }
    return values2; // So compilers don't complain
}

void
GaussLegendreValues::
getQuadrature(const int numOrdinates,
              std::vector<double>& weights,
              std::vector<Dim<1>::Vector>& ordinates) {
  if (numOrdinates == 1) {
    weights = {2.0};
    ordinates = {Dim<1>::Vector::zero};
  }
  else {
    // Get the quadrature values
    const auto& values = getValues(numOrdinates);
    
    // Get some indexing stuff and make sure we have an even quadrature order
    const int numValues = values.size() / 2;
    CHECK(numValues * 2 == numOrdinates);
    const int dataLength = 2;
    
    // Initialize the data
    weights.clear();
    ordinates.clear();
    ordinates.reserve(numOrdinates);
    weights.reserve(numOrdinates);
    
    // Add negative values in reverse so they are increasing
    for (auto j = 0; j < numValues; ++j) {
      const auto i = numValues - j - 1;
      ordinates.emplace_back(-values[0 + dataLength * i]);
      weights.push_back(values[1 + dataLength * i]);
    }
    
    // Add positive values 
    for (auto i = 0; i < numValues; ++i) {
      ordinates.emplace_back(values[0 + dataLength * i]);
      weights.push_back(values[1 + dataLength * i]);
    }
    
    // Make sure the ordinates got put in correctly
    CHECK(ordinates.size() == size_t(numOrdinates));
    CHECK(weights.size() == size_t(numOrdinates));
  }
}

void
GaussLegendreValues::
getQuadrature(const int numOrdinates,
              std::vector<double>& weights,
              std::vector<Dim<2>::Vector>& ordinates) {
  if (numOrdinates == 1) {
    weights = {2.0};
    ordinates = {Dim<2>::Vector::zero};
  }
  else {
    // Get the quadrature values
    const auto& values = getValues(numOrdinates);
    
    // Get some indexing stuff and make sure we have an even quadrature order
    const int numValues = values.size() / 2;
    CHECK(numValues * 2 == numOrdinates);
    const int dataLength = 2;
    
    // Initialize the data
    weights.clear();
    ordinates.clear();
    ordinates.reserve(numOrdinates);
    weights.reserve(numOrdinates);
    
    // Add negative values in reverse so they are increasing
    for (auto j = 0; j < numValues; ++j) {
      const auto i = numValues - j - 1;
      ordinates.emplace_back(-values[0 + dataLength * i], 0.0);
      weights.push_back(values[1 + dataLength * i]);
    }
    
    // Add positive values 
    for (auto i = 0; i < numValues; ++i) {
      ordinates.emplace_back(values[0 + dataLength * i], 0.0);
      weights.push_back(values[1 + dataLength * i]);
    }
    
    // Make sure the ordinates got put in correctly
    CHECK(ordinates.size() == size_t(numOrdinates));
    CHECK(weights.size() == size_t(numOrdinates));
  }
}

const std::vector<double> GaussLegendreValues::values2 =
  {5.7735026918962576451e-01, 1.0000000000000000000e0};
const std::vector<double> GaussLegendreValues::values4 =
  {3.3998104358485626480e-01, 6.5214515486254614263e-01,
   8.6113631159405257522e-01, 3.4785484513745385737e-01};
const std::vector<double> GaussLegendreValues::values8 =
  {1.8343464249564980494e-01, 3.6268378337836198297e-01,
   5.2553240991632898582e-01, 3.1370664587788728734e-01,
   7.9666647741362673959e-01, 2.2238103445337447054e-01,
   9.6028985649753623168e-01, 1.0122853629037625915e-01};
const std::vector<double> GaussLegendreValues::values16 =
  {9.5012509837637440185e-02, 1.8945061045506849629e-01,
   2.8160355077925891323e-01, 1.8260341504492358887e-01,
   4.5801677765722738634e-01, 1.6915651939500253819e-01,
   6.1787624440264374845e-01, 1.4959598881657673208e-01,
   7.5540440835500303390e-01, 1.2462897125553387205e-01,
   8.6563120238783174388e-01, 9.5158511682492784810e-02,
   9.4457502307323257608e-01, 6.2253523938647892863e-02,
   9.8940093499164993260e-01, 2.7152459411754094852e-02};
const std::vector<double> GaussLegendreValues::values32 =
  {4.8307665687738316235e-02, 9.6540088514727800567e-02,
   1.4447196158279649349e-01, 9.5638720079274859419e-02,
   2.3928736225213707454e-01, 9.3844399080804565639e-02,
   3.3186860228212764978e-01, 9.1173878695763884713e-02,
   4.2135127613063534536e-01, 8.7652093004403811143e-02,
   5.0689990893222939002e-01, 8.3311924226946755222e-02,
   5.8771575724076232904e-01, 7.8193895787070306472e-02,
   6.6304426693021520098e-01, 7.2345794108848506225e-02,
   7.3218211874028968039e-01, 6.5822222776361846838e-02,
   7.9448379596794240696e-01, 5.8684093478535547145e-02,
   8.4936761373256997013e-01, 5.0998059262376176196e-02,
   8.9632115576605212397e-01, 4.2835898022226680657e-02,
   9.3490607593773968917e-01, 3.4273862913021433103e-02,
   9.6476225558750643077e-01, 2.5392065309262059456e-02,
   9.8561151154526833540e-01, 1.6274394730905670605e-02,
   9.9726386184948156354e-01, 7.0186100094700966004e-03};
const std::vector<double> GaussLegendreValues::values64 =
  {2.4350292663424432509e-02, 4.8690957009139720383e-02,
   7.2993121787799039450e-02, 4.8575467441503426935e-02,
   1.2146281929612055447e-01, 4.8344762234802957170e-02,
   1.6964442042399281804e-01, 4.7999388596458307728e-02,
   2.1742364374000708415e-01, 4.7540165714830308662e-02,
   2.6468716220876741637e-01, 4.6968182816210017325e-02,
   3.1132287199021095616e-01, 4.6284796581314417296e-02,
   3.5722015833766811595e-01, 4.5491627927418144480e-02,
   4.0227015796399160370e-01, 4.4590558163756563060e-02,
   4.4636601725346408798e-01, 4.3583724529323453377e-02,
   4.8940314570705295748e-01, 4.2473515123653589007e-02,
   5.3127946401989454566e-01, 4.1262563242623528610e-02,
   5.7189564620263403428e-01, 3.9953741132720341387e-02,
   6.1115535517239325025e-01, 3.8550153178615629129e-02,
   6.4896547125465733986e-01, 3.7055128540240046040e-02,
   6.8523631305423324256e-01, 3.5472213256882383811e-02,
   7.1988185017161082685e-01, 3.3805161837141609392e-02,
   7.5281990726053189661e-01, 3.2057928354851553585e-02,
   7.8397235894334140761e-01, 3.0234657072402478868e-02,
   8.1326531512279755974e-01, 2.8339672614259483228e-02,
   8.4062929625258036275e-01, 2.6377469715054658672e-02,
   8.6599939815409281976e-01, 2.4352702568710873338e-02,
   8.8931544599511410585e-01, 2.2270173808383254159e-02,
   9.1052213707850280576e-01, 2.0134823153530209372e-02,
   9.2956917213193957582e-01, 1.7951715775697343085e-02,
   9.4641137485840281606e-01, 1.5726030476024719322e-02,
   9.6100879965205371892e-01, 1.3463047896718642598e-02,
   9.7332682778991096374e-01, 1.1168139460131128819e-02,
   9.8333625388462595693e-01, 8.8467598263639477230e-03,
   9.9101337147674432074e-01, 6.5044579689783628561e-03,
   9.9634011677195527935e-01, 4.1470332605624676353e-03,
   9.9930504173577213946e-01, 1.7832807216964329473e-03};
const std::vector<double> GaussLegendreValues::values128 =
  {1.2223698960615764198e-02, 2.4446180196262518211e-02,
   3.6663790968733493330e-02, 2.4431569097850045055e-02,
   6.1081969604139568104e-02, 2.4402355633849582093e-02,
   8.5463640504515498636e-02, 2.4358557264690625853e-02,
   1.0979423112764374667e-01, 2.4300200167971865323e-02,
   1.3405919946118778512e-01, 2.4227319222815248120e-02,
   1.5824404271422493400e-01, 2.4139957989019284998e-02,
   1.8233430598533718241e-01, 2.4038168681024052638e-02,
   2.0631559090207921715e-01, 2.3922012136703455672e-02,
   2.3017356422665998641e-01, 2.3791557781003400639e-02,
   2.5389396642269432086e-01, 2.3646883584447615144e-02,
   2.7746262017790440281e-01, 2.3488076016535913153e-02,
   3.0086543887767720267e-01, 2.3315229994062760122e-02,
   3.2408843502441337518e-01, 2.3128448824387027879e-02,
   3.4711772859763550843e-01, 2.2927844143686846920e-02,
   3.6993955534985902662e-01, 2.2713535850236461310e-02,
   3.9254027503326744274e-01, 2.2485652032744966872e-02,
   4.1490637955227501549e-01, 2.2244328893799765105e-02,
   4.3702450103710416294e-01, 2.1989710668460491434e-02,
   4.5888141983355219545e-01, 2.1721949538052075375e-02,
   4.8046407240417202586e-01, 2.1441205539208460137e-02,
   5.0175955913614446429e-01, 2.1147646468221348537e-02,
   5.2275515205117547845e-01, 2.0841447780751149114e-02,
   5.4343830241281036344e-01, 2.0522792486960069432e-02,
   5.6379664822661808391e-01, 2.0191871042130041181e-02,
   5.8381802162876308955e-01, 1.9848881232830862220e-02,
   6.0349045615854862420e-01, 1.9494028058706602823e-02,
   6.2280219391058491076e-01, 1.9127523609950945487e-02,
   6.4174169256230755715e-01, 1.8749586940544708651e-02,
   6.6029763227264605211e-01, 1.8360443937331343221e-02,
   6.7845892244771925937e-01, 1.7960327185008685940e-02,
   6.9621470836951433239e-01, 1.7549475827117704649e-02,
   7.1355437768358741334e-01, 1.7128135423111376831e-02,
   7.3046756674190880647e-01, 1.6696557801589204589e-02,
   7.4694416679706198117e-01, 1.6255000909785187052e-02,
   7.6297433004409472278e-01, 1.5803728659399346859e-02,
   7.7854847550641196685e-01, 1.5343010768865144086e-02,
   7.9365729476219329024e-01, 1.4873122602147314252e-02,
   8.0829175750791366012e-01, 1.4394345004166846177e-02,
   8.2244311695564384246e-01, 1.3906964132951985244e-02,
   8.3610291506090684712e-01, 1.3411271288616332314e-02,
   8.4926298757796896916e-01, 1.2907562739267347220e-02,
   8.6191546893954846059e-01, 1.2396139543950922969e-02,
   8.7405279695803179870e-01, 1.1877307372740279576e-02,
   8.8566771734539721741e-01, 1.1351376324080416693e-02,
   8.9675328804915818439e-01, 1.0818660739503076248e-02,
   9.0730288340175681392e-01, 1.0279479015832157133e-02,
   9.1731019808096053704e-01, 9.7341534150068058635e-03,
   9.2676925087894784333e-01, 9.1830098716608743345e-03,
   9.3567438827791637578e-01, 8.6263777986167497050e-03,
   9.4402028783022018212e-01, 8.0645898904860579729e-03,
   9.5180196134126438622e-01, 7.4979819256347286877e-03,
   9.5901475785369992810e-01, 6.9268925668988135634e-03,
   9.6565436643196526865e-01, 6.3516631617071887872e-03,
   9.7171681874713658090e-01, 5.7726375428656985893e-03,
   9.7719849146390738717e-01, 5.1901618326763302051e-03,
   9.8209610843571853602e-01, 4.6045842567029551183e-03,
   9.8640674272458620887e-01, 4.0162549837386423132e-03,
   9.9012781849173438334e-01, 3.4255260409102157743e-03,
   9.9325711290021293530e-01, 2.8327514714579910953e-03,
   9.9579275853498118686e-01, 2.2382884309626187436e-03,
   9.9773324862551401988e-01, 1.6425030186690295388e-03,
   9.9907745997737589501e-01, 1.0458126793403487793e-03,
   9.9982488794713191447e-01, 4.4938096029209037639e-04};
const std::vector<double> GaussLegendreValues::values256 =
  {6.1239123751895295012e-03, 1.2247671640289755904e-02,
   1.8370818478813665118e-02, 1.2245834369747920142e-02,
   3.0614968779979029366e-02, 1.2242160104272800770e-02,
   4.2854526536379098381e-02, 1.2236649395040158109e-02,
   5.5087655694633984105e-02, 1.2229303068710278904e-02,
   6.7312521165716400242e-02, 1.2220122227303969192e-02,
   7.9527289100232965903e-02, 1.2209108248037240408e-02,
   9.1730127163519552031e-02, 1.2196262783114713518e-02,
   1.0391920481050940364e-01, 1.2181587759481772174e-02,
   1.1609269356033280494e-01, 1.2165085378535502061e-02,
   1.2824876727060709474e-01, 1.2146758115794459816e-02,
   1.4038560241137588591e-01, 1.2126608720527321035e-02,
   1.5250137833865639537e-01, 1.2104640215340463098e-02,
   1.6459427756755384983e-01, 1.2080855895724544656e-02,
   1.7666248604490199740e-01, 1.2055259329560149814e-02,
   1.8870419342138882646e-01, 1.2027854356582571161e-02,
   2.0071759332312667007e-01, 1.1998645087805811935e-02,
   2.1270088362262595794e-01, 1.1967635904905893729e-02,
   2.2465226670913196715e-01, 1.1934831459563562256e-02,
   2.3656994975828401848e-01, 1.1900236672766489754e-02,
   2.4845214500105666683e-01, 1.1863856734071078732e-02,
   2.6029706999194254198e-01, 1.1825697100823977771e-02,
   2.7210294787633660951e-01, 1.1785763497343426182e-02,
   2.8386800765708174180e-01, 1.1744061914060550305e-02,
   2.9559048446013561456e-01, 1.1700598606620740288e-02,
   3.0726861979931907626e-01, 1.1655380094945242121e-02,
   3.1890066184010627563e-01, 1.1608413162253105722e-02,
   3.3048486566241697623e-01, 1.1559704854043635773e-02,
   3.4201949352237163648e-01, 1.1509262477039497959e-02,
   3.5350281511296998954e-01, 1.1457093598090639152e-02,
   3.6493310782365401853e-01, 1.1403206043039185965e-02,
   3.7630865699871639028e-01, 1.1347607895545491942e-02,
   3.8762775619451558364e-01, 1.1290307495875509508e-02,
   3.9888870743545912771e-01, 1.1231313439649668573e-02,
   4.1008982146871655001e-01, 1.1170634576553449463e-02,
   4.2122941801762382498e-01, 1.1108280009009843630e-02,
   4.3230582603374130995e-01, 1.1044259090813901264e-02,
   4.4331738394752735722e-01, 1.0978581425729570638e-02,
   4.5426243991758999877e-01, 1.0911256866049039701e-02,
   4.6513935207847931365e-01, 1.0842295511114795995e-02,
   4.7594648878698330639e-01, 1.0771707705804626637e-02,
   4.8668222886689035010e-01, 1.0699504038979785603e-02,
   4.9734496185218147712e-01, 1.0625695341896561134e-02,
   5.0793308822861603623e-01, 1.0550292686581481518e-02,
   5.1844501967367447622e-01, 1.0473307384170403004e-02,
   5.2887917929482226195e-01, 1.0394750983211728997e-02,
   5.3923400186605918113e-01, 1.0314635267934015068e-02,
   5.4950793406271855704e-01, 1.0232972256478219657e-02,
   5.5969943469448114514e-01, 1.0149774199094865655e-02,
   5.6980697493656875906e-01, 1.0065053576306383309e-02,
   5.7982903855908294492e-01, 9.9788230970349101247e-03,
   5.8976412215445430079e-01, 9.8910956966958286026e-03,
   5.9961073536296832173e-01, 9.8018845352573278255e-03,
   6.0936740109633393952e-01, 9.7112029952662799642e-03,
   6.1903265575926121943e-01, 9.6190646798407278572e-03,
   6.2860504946901497543e-01, 9.5254834106292848118e-03,
   6.3808314627291136867e-01, 9.4304732257377527474e-03,
   6.4746552436372486262e-01, 9.3340483776232697125e-03,
   6.5675077629297322189e-01, 9.2362233309563026874e-03,
   6.6593750918204855991e-01, 9.1370127604508064020e-03,
   6.7502434493116276386e-01, 9.0364315486628736802e-03,
   6.8400992042607595312e-01, 8.9344947837582075484e-03,
   6.9289288774257696011e-01, 8.8312177572487500253e-03,
   7.0167191434868515941e-01, 8.7266159616988071403e-03,
   7.1034568330454331339e-01, 8.6207050884010143054e-03,
   7.1891289345997144837e-01, 8.5135010250224906938e-03,
   7.2737225964965212659e-01, 8.4050198532215357562e-03,
   7.3572251288591783462e-01, 8.2952778462352254252e-03,
   7.4396240054911156846e-01, 8.1842914664382699356e-03,
   7.5209068657549205959e-01, 8.0720773628734995009e-03,
   7.6010615164265545494e-01, 7.9586523687543483536e-03,
   7.6800759335244563598e-01, 7.8440334989397118668e-03,
   7.7579382641132573913e-01, 7.7282379473815556311e-03,
   7.8346368280818382075e-01, 7.6112830845456594616e-03,
   7.9101601198954599455e-01, 7.4931864548058833586e-03,
   7.9844968103217075878e-01, 7.3739657738123464376e-03,
   8.0576357481299862326e-01, 7.2536389258339137838e-03,
   8.1295659617643154314e-01, 7.1322239610753900717e-03,
   8.2002766609891706740e-01, 7.0097390929698226212e-03,
   8.2697572385081251429e-01, 6.8862026954463203467e-03,
   8.3379972715550489435e-01, 6.7616333001737987809e-03,
   8.4049865234576271390e-01, 6.6360495937810650446e-03,
   8.4707149451729620719e-01, 6.5094704150536602678e-03,
   8.5351726767950296507e-01, 6.3819147521078805704e-03,
   8.5983500490337635070e-01, 6.2534017395424012721e-03,
   8.6602375846655451930e-01, 6.1239506555679325424e-03,
   8.7208259999548828913e-01, 5.9935809191153382211e-03,
   8.7801062060470654399e-01, 5.8623120869226530607e-03,
   8.8380693103315828486e-01, 5.7301638506014371774e-03,
   8.8947066177761088883e-01, 5.5971560336829100776e-03,
   8.9500096322308457744e-01, 5.4633085886443102776e-03,
   9.0039700577030354477e-01, 5.3286415939159303171e-03,
   9.0565797996014464708e-01, 5.1931752508692809303e-03,
   9.1078309659506501189e-01, 5.0569298807868423876e-03,
   9.1577158685749038453e-01, 4.9199259218138656696e-03,
   9.2062270242514649551e-01, 4.7821839258926913729e-03,
   9.2533571558331620287e-01, 4.6437245556800603140e-03,
   9.2990991933400564118e-01, 4.5045685814478970686e-03,
   9.3434462750200309429e-01, 4.3647368779680566816e-03,
   9.3863917483781480498e-01, 4.2242504213815362724e-03,
   9.4279291711746244318e-01, 4.0831302860526684086e-03,
   9.4680523123912748137e-01, 3.9413976414088336277e-03,
   9.5067551531662827636e-01, 3.7990737487662579981e-03,
   9.5440318876971624176e-01, 3.6561799581425021694e-03,
   9.5798769241117812937e-01, 3.5127377050563073310e-03,
   9.6142848853073214401e-01, 3.3687685073155510120e-03,
   9.6472506097570643093e-01, 3.2242939617941981570e-03,
   9.6787691522848945491e-01, 3.0793357411993375832e-03,
   9.7088357848074302932e-01, 2.9339155908297166460e-03,
   9.7374459970437040527e-01, 2.7880553253277068806e-03,
   9.7645954971923415562e-01, 2.6417768254274905641e-03,
   9.7902802125762203882e-01, 2.4951020347037068508e-03,
   9.8144962902546440577e-01, 2.3480529563273120170e-03,
   9.8372400976031549617e-01, 2.2006516498399104997e-03,
   9.8585082228612595648e-01, 2.0529202279661431745e-03,
   9.8782974756486060892e-01, 1.9048808534997184044e-03,
   9.8966048874506521832e-01, 1.7565557363307299936e-03,
   9.9134277120758308692e-01, 1.6079671307493272424e-03,
   9.9287634260882211714e-01, 1.4591373333107332011e-03,
   9.9426097292240966496e-01, 1.3100886819025044578e-03,
   9.9549645448109635659e-01, 1.1608435575677247240e-03,
   9.9658260202338154043e-01, 1.0114243932084404526e-03,
   9.9751925275672082756e-01, 8.6185370142008903781e-04,
   9.9830626647300644406e-01, 7.1215416347332066691e-04,
   9.9894352584340885656e-01, 5.6234895403140980282e-04,
   9.9943093746626140824e-01, 4.1246325442617632843e-04,
   9.9976843740926318610e-01, 2.6253494429644590629e-04,
   9.9995605001899223073e-01, 1.1278901782227217551e-04};
const std::vector<double> GaussLegendreValues::values512 =
  {3.0649621851593961529e-03, 6.1299051754057857592e-03,
   9.1947713864329108047e-03, 6.1296748380364986664e-03,
   1.5324235084898185525e-02, 6.1292141719530834395e-03,
   2.1453122959774874514e-02, 6.1285231944655327693e-03,
   2.7581204711919784062e-02, 6.1276019315380226385e-03,
   3.3708250072480595123e-02, 6.1264504177879366912e-03,
   3.9834028811548447683e-02, 6.1250686964845654507e-03,
   4.5958310746809061779e-02, 6.1234568195474804312e-03,
   5.2080865752192070113e-02, 6.1216148475445832082e-03,
   5.8201463766518237239e-02, 6.1195428496898295184e-03,
   6.4319874802144240405e-02, 6.1172409038406284754e-03,
   7.0435868953604687199e-02, 6.1147090964949169991e-03,
   7.6549216406251045292e-02, 6.1119475227879095685e-03,
   8.2659687444887159628e-02, 6.1089562864885234199e-03,
   8.8767052462401032609e-02, 6.1057354999954793256e-03,
   9.4871081968392542891e-02, 6.1022852843330780982e-03,
   1.0097154659779677863e-01, 6.0986057691466529805e-03,
   1.0706821711950266111e-01, 6.0946970926976980917e-03,
   1.1316086444496653494e-01, 6.0905594018586731119e-03,
   1.1924925963682040116e-01, 6.0861928521074844015e-03,
   1.2533317391747446969e-01, 6.0815976075216427621e-03,
   1.3141237867771370801e-01, 6.0767738407720980584e-03,
   1.3748664548528806302e-01, 6.0717217331167509334e-03,
   1.4355574609349603317e-01, 6.0664414743936418599e-03,
   1.4961945244976126852e-01, 6.0609332630138177842e-03,
   1.5567753670420187625e-01, 6.0551973059538766317e-03,
   1.6172977121819210980e-01, 6.0492338187481899521e-03,
   1.6777592857291611981e-01, 6.0430430254808039979e-03,
   1.7381578157791344550e-01, 6.0366251587770195405e-03,
   1.7984910327961592534e-01, 6.0299804597946507400e-03,
   1.8587566696987570627e-01, 6.0231091782149633973e-03,
   1.9189524619448403102e-01, 6.0160115722332929282e-03,
   1.9790761476168048340e-01, 6.0086879085493424136e-03,
   2.0391254675065237177e-01, 6.0011384623571610896e-03,
   2.0990981652002393149e-01, 5.9933635173348036527e-03,
   2.1589919871633502719e-01, 5.9853633656336707716e-03,
   2.2188046828250903625e-01, 5.9771383078675312031e-03,
   2.2785340046630959551e-01, 5.9686886531012259272e-03,
   2.3381777082878589318e-01, 5.9600147188390547234e-03,
   2.3977335525270618879e-01, 5.9511168310128456268e-03,
   2.4571992995097924421e-01, 5.9419953239697077108e-03,
   2.5165727147506334932e-01, 5.9326505404594676575e-03,
   2.5758515672336262628e-01, 5.9230828316217905873e-03,
   2.6350336294961029706e-01, 5.9132925569729856313e-03,
   2.6941166777123859903e-01, 5.9032800843924967444e-03,
   2.7530984917773503422e-01, 5.8930457901090792634e-03,
   2.8119768553898463830e-01, 5.8825900586866627325e-03,
   2.8707495561359795560e-01, 5.8719132830099005256e-03,
   2.9294143855722440749e-01, 5.8610158642694068094e-03,
   2.9879691393085074159e-01, 5.8498982119466814015e-03,
   3.0464116170908425001e-01, 5.8385607437987230902e-03,
   3.1047396228842044539e-01, 5.8270038858423319934e-03,
   3.1629509649549488407e-01, 5.8152280723381015486e-03,
   3.2210434559531882630e-01, 5.8032337457741007325e-03,
   3.2790149129949842406e-01, 5.7910213568492471258e-03,
   3.3368631577443712757e-01, 5.7785913644563714469e-03,
   3.3945860164952100247e-01, 5.7659442356649741911e-03,
   3.4521813202528664978e-01, 5.7530804457036750229e-03,
   3.5096469048157142204e-01, 5.7400004779423555815e-03,
   3.5669806108564562917e-01, 5.7267048238739963700e-03,
   3.6241802840032642859e-01, 5.7131939830962084111e-03,
   3.6812437749207309466e-01, 5.6994684632924603630e-03,
   3.7381689393906336318e-01, 5.6855287802130018011e-03,
   3.7949536383925054770e-01, 5.6713754576554833824e-03,
   3.8515957381840112460e-01, 5.6570090274452746203e-03,
   3.9080931103811248515e-01, 5.6424300294154800103e-03,
   3.9644436320381055312e-01, 5.6276390113866542567e-03,
   4.0206451857272696754e-01, 5.6126365291462173627e-03,
   4.0766956596185553077e-01, 5.5974231464275703576e-03,
   4.1325929475588762292e-01, 5.5819994348889124461e-03,
   4.1883349491512628455e-01, 5.5663659740917603748e-03,
   4.2439195698337867005e-01, 5.5505233514791708236e-03,
   4.2993447209582657541e-01, 5.5344721623536666407e-03,
   4.3546083198687474434e-01, 5.5182130098548677502e-03,
   4.4097082899797665813e-01, 5.5017465049368275724e-03,
   4.4646425608543751494e-01, 5.4850732663450758090e-03,
   4.5194090682819410545e-01, 5.4681939205933684566e-03,
   4.5740057543557129250e-01, 5.4511091019401459197e-03,
   4.6284305675501480328e-01, 5.4338194523647001110e-03,
   4.6826814627980004343e-01, 5.4163256215430514317e-03,
   4.7367564015671664352e-01, 5.3986282668235365401e-03,
   4.7906533519372844899e-01, 5.3807280532021078252e-03,
   4.8443702886760866589e-01, 5.3626256532973455128e-03,
   4.8979051933154987531e-01, 5.3443217473251833447e-03,
   4.9512560542274863085e-01, 5.3258170230733487788e-03,
   5.0044208666996435375e-01, 5.3071121758755186716e-03,
   5.0573976330105224198e-01, 5.2882079085851914147e-03,
   5.1101843625046991011e-01, 5.2691049315492765055e-03,
   5.1627790716675747776e-01, 5.2498039625814025460e-03,
   5.2151797841999082581e-01, 5.2303057269349446720e-03,
   5.2673845310920774012e-01, 5.2106109572757724262e-03,
   5.3193913506980666376e-01, 5.1907203936547190996e-03,
   5.3711982888091777977e-01, 5.1706347834797735753e-03,
   5.4228033987274614743e-01, 5.1503548814879957195e-03,
   5.4742047413388661617e-01, 5.1298814497171563759e-03,
   5.5254003851861024216e-01, 5.1092152574771030282e-03,
   5.5763884065412193394e-01, 5.0883570813208522065e-03,
   5.6271668894778905413e-01, 5.0673077050154097257e-03,
   5.6777339259434070593e-01, 5.0460679195123198490e-03,
   5.7280876158303743356e-01, 5.0246385229179444874e-03,
   5.7782260670481106746e-01, 5.0030203204634735478e-03,
   5.8281473955937444588e-01, 4.9812141244746675595e-03,
   5.8778497256230074564e-01, 4.9592207543413337152e-03,
   5.9273311895207215623e-01, 4.9370410364865364724e-03,
   5.9765899279709763216e-01, 4.9146758043355438745e-03,
   6.0256240900269946004e-01, 4.8921258982845107556e-03,
   6.0744318331806837780e-01, 4.8693921656689000083e-03,
   6.1230113234318698466e-01, 4.8464754607316430994e-03,
   6.1713607353572118180e-01, 4.8233766445910410308e-03,
   6.2194782521787938463e-01, 4.8000965852084069517e-03,
   6.2673620658323924910e-01, 4.7766361573554516371e-03,
   6.3150103770354165535e-01, 4.7529962425814130595e-03,
   6.3624213953545169356e-01, 4.7291777291799312876e-03,
   6.4095933392728639782e-01, 4.7051815121556699580e-03,
   6.4565244362570897533e-01, 4.6810084931906855725e-03,
   6.5032129228238927931e-01, 4.6566595806105458870e-03,
   6.5496570446063027537e-01, 4.6321356893501986622e-03,
   6.5958550564196025237e-01, 4.6074377409195920619e-03,
   6.6418052223269053000e-01, 4.5825666633690479878e-03,
   6.6875058157043841678e-01, 4.5575233912543896536e-03,
   6.7329551193061517318e-01, 4.5323088656018247089e-03,
   6.7781514253287873634e-01, 4.5069240338725852313e-03,
   6.8230930354755096360e-01, 4.4813698499273259161e-03,
   6.8677782610199915404e-01, 4.4556472739902818017e-03,
   6.9122054228698160796e-01, 4.4297572726131868769e-03,
   6.9563728516295698599e-01, 4.4037008186389549258e-03,
   7.0002788876635723079e-01, 4.3774788911651239763e-03,
   7.0439218811582381554e-01, 4.3510924755070657235e-03,
   7.0873001921840708485e-01, 4.3245425631609613132e-03,
   7.1304121907572845534e-01, 4.2978301517665448748e-03,
   7.1732562569010524412e-01, 4.2709562450696162035e-03,
   7.2158307807063789512e-01, 4.2439218528843240023e-03,
   7.2581341623925937456e-01, 4.2167279910552210986e-03,
   7.3001648123674650824e-01, 4.1893756814190930635e-03,
   7.3419211512869303465e-01, 4.1618659517665616659e-03,
   7.3834016101144414969e-01, 4.1341998358034646067e-03,
   7.4246046301799231972e-01, 4.1063783731120129818e-03,
   7.4655286632383414169e-01, 4.0784026091117279353e-03,
   7.5061721715278803003e-01, 4.0502735950201579699e-03,
   7.5465336278277251181e-01, 4.0219923878133783908e-03,
   7.5866115155154491307e-01, 3.9935600501862743674e-03,
   7.6264043286240022060e-01, 3.9649776505126091054e-03,
   7.6659105718982990509e-01, 3.9362462628048786290e-03,
   7.7051287608514049300e-01, 3.9073669666739546834e-03,
   7.7440574218203167601e-01, 3.8783408472885172720e-03,
   7.7826950920213374846e-01, 3.8491689953342783541e-03,
   7.8210403196050416472e-01, 3.8198525069729982349e-03,
   7.8590916637108300996e-01, 3.7903924838012961884e-03,
   7.8968476945210717919e-01, 3.7607900328092568595e-03,
   7.9343069933148306144e-01, 3.7310462663388340022e-03,
   7.9714681525211752676e-01, 3.7011623020420531167e-03,
   8.0083297757720701619e-01, 3.6711392628390145554e-03,
   8.0448904779548453552e-01, 3.6409782768756986764e-03,
   8.0811488852642435609e-01, 3.6106804774815746301e-03,
   8.1171036352540422664e-01, 3.5802470031270143714e-03,
   8.1527533768882490267e-01, 3.5496789973805134987e-03,
   8.1880967705918680055e-01, 3.5189776088657205262e-03,
   8.2231324883012358588e-01, 3.4881439912182762046e-03,
   8.2578592135139250684e-01, 3.4571793030424645128e-03,
   8.2922756413382128508e-01, 3.4260847078676769484e-03,
   8.3263804785421137815e-01, 3.3948613741046917538e-03,
   8.3601724436019742944e-01, 3.3635104750017697209e-03,
   8.3936502667506272275e-01, 3.3320331886005682237e-03,
   8.4268126900251046083e-01, 3.3004306976918751358e-03,
   8.4596584673139068838e-01, 3.2687041897711642972e-03,
   8.4921863644038268202e-01, 3.2368548569939741987e-03,
   8.5243951590263263128e-01, 3.2048838961311115628e-03,
   8.5562836409034643626e-01, 3.1727925085236815030e-03,
   8.5878506117933744951e-01, 3.1405819000379459532e-03,
   8.6190948855352899111e-01, 3.1082532810200120618e-03,
   8.6500152880941146790e-01, 3.0758078662503522550e-03,
   8.6806106576045392928e-01, 3.0432468748981576781e-03,
   8.7108798444146989389e-01, 3.0105715304755267298e-03,
   8.7408217111293728300e-01, 2.9777830607914904130e-03,
   8.7704351326527229854e-01, 2.9448826979058762279e-03,
   8.7997189962305708483e-01, 2.9118716780830123435e-03,
   8.8286722014922101550e-01, 2.8787512417452737869e-03,
   8.8572936604917544824e-01, 2.8455226334264723965e-03,
   8.8855822977490179214e-01, 2.8121871017250922922e-03,
   8.9135370502899273402e-01, 2.7787458992573726197e-03,
   8.9411568676864647187e-01, 2.7452002826102393336e-03,
   8.9684407120961380525e-01, 2.7115515122940877888e-03,
   8.9953875583009793455e-01, 2.6778008526954179164e-03,
   9.0219963937460682236e-01, 2.6439495720293237640e-03,
   9.0482662185775797238e-01, 2.6099989422918391897e-03,
   9.0741960456803548277e-01, 2.5759502392121415000e-03,
   9.0997849007149923296e-01, 2.5418047422046148319e-03,
   9.1250318221544606434e-01, 2.5075637343207750815e-03,
   9.1499358613202281753e-01, 2.4732285022010581904e-03,
   9.1744960824179109027e-01, 2.4388003360264736029e-03,
   9.1987115625724358221e-01, 2.4042805294701247170e-03,
   9.2225813918627189428e-01, 2.3696703796485981536e-03,
   9.2461046733558565265e-01, 2.3349711870732236769e-03,
   9.2692805231408282858e-01, 2.3001842556012066043e-03,
   9.2921080703617112775e-01, 2.2653108923866345475e-03,
   9.3145864572504032428e-01, 2.2303524078313603368e-03,
   9.3367148391588541648e-01, 2.1953101155357629824e-03,
   9.3584923845908048340e-01, 2.1601853322493885355e-03,
   9.3799182752330312299e-01, 2.1249793778214727179e-03,
   9.4009917059860935448e-01, 2.0896935751513471948e-03,
   9.4217118849945886972e-01, 2.0543292501387313744e-03,
   9.4420780336769051982e-01, 2.0188877316339116256e-03,
   9.4620893867544792553e-01, 1.9833703513878098109e-03,
   9.4817451922805510157e-01, 1.9477784440019430461e-03,
   9.5010447116684198719e-01, 1.9121133468782766037e-03,
   9.5199872197191977698e-01, 1.8763764001689718922e-03,
   9.5385720046490594799e-01, 1.8405689467260314558e-03,
   9.5567983681159888119e-01, 1.8046923320508429542e-03,
   9.5746656252460197723e-01, 1.7687479042436241016e-03,
   9.5921731046589716847e-01, 1.7327370139527705643e-03,
   9.6093201484936773117e-01, 1.6966610143241088446e-03,
   9.6261061124327030396e-01, 1.6605212609500562073e-03,
   9.6425303657265602064e-01, 1.6243191118186897474e-03,
   9.6585922912174066745e-01, 1.5880559272627267421e-03,
   9.6742912853622377734e-01, 1.5517330699084184929e-03,
   9.6896267582555657566e-01, 1.5153519046243599371e-03,
   9.7045981336515869446e-01, 1.4789137984702174060e-03,
   9.7192048489858357452e-01, 1.4424201206453770260e-03,
   9.7334463553963247735e-01, 1.4058722424375164226e-03,
   9.7473221177441703157e-01, 1.3692715371711025869e-03,
   9.7608316146337024168e-01, 1.3326193801558190401e-03,
   9.7739743384320588997e-01, 1.2959171486349257825e-03,
   9.7867497952882626643e-01, 1.2591662217335559931e-03,
   9.7991575051517816567e-01, 1.2223679804069540809e-03,
   9.8111970017905709473e-01, 1.1855238073886605549e-03,
   9.8228678328085964192e-01, 1.1486350871386503607e-03,
   9.8341695596628396407e-01, 1.1117032057914329650e-03,
   9.8451017576797835907e-01, 1.0747295511041247428e-03,
   9.8556640160713790247e-01, 1.0377155124045074301e-03,
   9.8658559379504914296e-01, 1.0006624805390909706e-03,
   9.8756771403458287298e-01, 9.6357184782120567985e-04,
   9.8851272542163502001e-01, 9.2644500797915826975e-04,
   9.8942059244651574538e-01, 8.8928335610450053720e-04,
   9.9029128099528689621e-01, 8.5208828860048094028e-04,
   9.9112475835104804155e-01, 8.1486120313078199656e-04,
   9.9192099319517145002e-01, 7.7760349856869724380e-04,
   9.9267995560848655735e-01, 7.4031657494698189629e-04,
   9.9340161707241476579e-01, 7.0300183340874114339e-04,
   9.9408595047005587937e-01, 6.6566067615993434094e-04,
   9.9473293008722822250e-01, 6.2829450642443583909e-04,
   9.9534253161346571515e-01, 5.9090472840322301624e-04,
   9.9591473214297725670e-01, 5.5349274724038946478e-04,
   9.9644951017557740228e-01, 5.1605996900076743710e-04,
   9.9694684561760388044e-01, 4.7860780066795090669e-04,
   9.9740671978284983216e-01, 4.4113765017954056365e-04,
   9.9782911539356284660e-01, 4.0365092653331987974e-04,
   9.9821401658161279539e-01, 3.6614904003562685301e-04,
   9.9856140889003972756e-01, 3.2863340285233341625e-04,
   9.9887127927544942465e-01, 2.9110543025148851253e-04,
   9.9914361611237823825e-01, 2.5356654357058651359e-04,
   9.9937840920259925145e-01, 2.1601817797699085834e-04,
   9.9957564979831085559e-01, 1.7846180554595329461e-04,
   9.9973533067104266258e-01, 1.4089901738819849301e-04,
   9.9985744636997943854e-01, 1.0333190349691323630e-04,
   9.9994199460684565364e-01, 6.5765731659240195831e-05,
   9.9998899098438186799e-01, 2.8252637373934692039e-05};
const std::vector<double> GaussLegendreValues::values1024 =
  {1.5332313560626384065e-03, 3.0664603092439082116e-03,
   4.5996796509132604743e-03, 3.0664314747171934850e-03,
   7.6660846940754867628e-03, 3.0663738059349007324e-03,
   1.0732417651542280333e-02, 3.0662873034393008057e-03,
   1.3798649689984440154e-02, 3.0661719680437936084e-03,
   1.6864751977021726545e-02, 3.0660278008329004478e-03,
   1.9930695681493977691e-02, 3.0658548031622538364e-03,
   2.2996451973732214686e-02, 3.0656529766585847451e-03,
   2.6061992025829732558e-02, 3.0654223232197073064e-03,
   2.9127287011913174719e-02, 3.0651628450145009692e-03,
   3.2192308108413588295e-02, 3.0648745444828901040e-03,
   3.5257026494337457792e-02, 3.0645574243358210601e-03,
   3.8321413351537714538e-02, 3.0642114875552366740e-03,
   4.1385439864984719363e-02, 3.0638367373940482296e-03,
   4.4449077223037215969e-02, 3.0634331773761048702e-03,
   4.7512296617713252429e-02, 3.0630008112961604636e-03,
   5.0575069244961068282e-02, 3.0625396432198379187e-03,
   5.3637366304929944678e-02, 3.0620496774835909559e-03,
   5.6699159002241015007e-02, 3.0615309186946633309e-03,
   5.9760418546258033485e-02, 3.0609833717310455112e-03,
   6.2821116151358099149e-02, 3.0604070417414288080e-03,
   6.5881223037202332700e-02, 3.0598019341451569616e-03,
   6.8940710429006503669e-02, 3.0591680546321751827e-03,
   7.1999549557811605345e-02, 3.0585054091629766484e-03,
   7.5057711660754374928e-02, 3.0578140039685464546e-03,
   7.8115167981337756370e-02, 3.0570938455503030247e-03,
   8.1171889769701303340e-02, 3.0563449406800369760e-03,
   8.4227848282891519798e-02, 3.0555672963998474425e-03,
   8.7283014785132135609e-02, 3.0547609200220758572e-03,
   9.0337360548094314680e-02, 3.0539258191292371925e-03,
   9.3390856851166793053e-02, 3.0530620015739486603e-03,
   9.6443474981725944445e-02, 3.0521694754788558725e-03,
   9.9495186235405770664e-02, 3.0512482492365564620e-03,
   1.0254596191636781439e-01, 3.0502983315095211654e-03,
   1.0559577333757099174e-01, 3.0493197312300123682e-03,
   1.0864459182104134218e-01, 3.0483124576000001133e-03,
   1.1169238869814169307e-01, 3.0472765200910755724e-03,
   1.1473913530984123652e-01, 3.0462119284443619832e-03,
   1.1778480300698501585e-01, 3.0451186926704230517e-03,
   1.2082936315056331919e-01, 3.0439968230491688209e-03,
   1.2387278711198097773e-01, 3.0428463301297590067e-03,
   1.2691504627332656597e-01, 3.0416672247305038022e-03,
   1.2995611202764150157e-01, 3.0404595179387621506e-03,
   1.3299595577918904218e-01, 3.0392232211108374895e-03,
   1.3603454894372317672e-01, 3.0379583458718709643e-03,
   1.3907186294875740870e-01, 3.0366649041157321154e-03,
   1.4210786923383342885e-01, 3.0353429080049070377e-03,
   1.4514253925078967473e-01, 3.0339923699703840143e-03,
   1.4817584446402977469e-01, 3.0326133027115366252e-03,
   1.5120775635079087364e-01, 3.0312057191960043331e-03,
   1.5423824640141183819e-01, 3.0297696326595705460e-03,
   1.5726728611960133861e-01, 3.0283050566060381583e-03,
   1.6029484702270580496e-01, 3.0268120048071025721e-03,
   1.6332090064197725514e-01, 3.0252904913022221991e-03,
   1.6634541852284099205e-01, 3.0237405303984864452e-03,
   1.6936837222516316753e-01, 3.0221621366704811777e-03,
   1.7238973332351821055e-01, 3.0205553249601516777e-03,
   1.7540947340745611699e-01, 3.0189201103766630786e-03,
   1.7842756408176959871e-01, 3.0172565082962582916e-03,
   1.8144397696676108925e-01, 3.0155645343621134196e-03,
   1.8445868369850960363e-01, 3.0138442044841906616e-03,
   1.8747165592913744990e-01, 3.0120955348390887083e-03,
   1.9048286532707678978e-01, 3.0103185418698906302e-03,
   1.9349228357733604592e-01, 3.0085132422860092601e-03,
   1.9649988238176615332e-01, 3.0066796530630300711e-03,
   1.9950563345932665238e-01, 3.0048177914425515522e-03,
   2.0250950854635162104e-01, 3.0029276749320230818e-03,
   2.0551147939681544356e-01, 3.0010093213045803019e-03,
   2.0851151778259841347e-01, 2.9990627485988779939e-03,
   2.1150959549375216805e-01, 2.9970879751189204574e-03,
   2.1450568433876495206e-01, 2.9950850194338893942e-03,
   2.1749975614482670799e-01, 2.9930539003779692986e-03,
   2.2049178275809399053e-01, 2.9909946370501703558e-03,
   2.2348173604395470268e-01, 2.9889072488141488505e-03,
   2.2646958788729265103e-01, 2.9867917552980250862e-03,
   2.2945531019275191766e-01, 2.9846481763941988184e-03,
   2.3243887488500104630e-01, 2.9824765322591622023e-03,
   2.3542025390899704016e-01, 2.9802768433133102578e-03,
   2.3839941923024916903e-01, 2.9780491302407488518e-03,
   2.4137634283508258301e-01, 2.9757934139891002022e-03,
   2.4435099673090173066e-01, 2.9735097157693059029e-03,
   2.4732335294645357879e-01, 2.9711980570554274732e-03,
   2.5029338353209063169e-01, 2.9688584595844444332e-03,
   2.5326106056003374709e-01, 2.9664909453560499065e-03,
   2.5622635612463474654e-01, 2.9640955366324437529e-03,
   2.5918924234263881774e-01, 2.9616722559381232326e-03,
   2.6214969135344670615e-01, 2.9592211260596712038e-03,
   2.6510767531937669376e-01, 2.9567421700455418562e-03,
   2.6806316642592636218e-01, 2.9542354112058439816e-03,
   2.7101613688203413791e-01, 2.9517008731121217846e-03,
   2.7396655892034061708e-01, 2.9491385795971332349e-03,
   2.7691440479744966743e-01, 2.9465485547546259626e-03,
   2.7985964679418930485e-01, 2.9439308229391107008e-03,
   2.8280225721587234219e-01, 2.9412854087656322747e-03,
   2.8574220839255680784e-01, 2.9386123371095381419e-03,
   2.8867947267930613160e-01, 2.9359116331062444843e-03,
   2.9161402245644909544e-01, 2.9331833221509998553e-03,
   2.9454583012983954667e-01, 2.9304274298986463829e-03,
   2.9747486813111587109e-01, 2.9276439822633785324e-03,
   3.0040110891796022373e-01, 2.9248330054184994302e-03,
   3.0332452497435751460e-01, 2.9219945257961747508e-03,
   3.0624508881085414723e-01, 2.9191285700871841706e-03,
   3.0916277296481650732e-01, 2.9162351652406703884e-03,
   3.1207755000068919933e-01, 2.9133143384638857180e-03,
   3.1498939251025302832e-01, 2.9103661172219362530e-03,
   3.1789827311288272483e-01, 2.9073905292375236068e-03,
   3.2080416445580441026e-01, 2.9043876024906842307e-03,
   3.2370703921435280037e-01, 2.9013573652185263121e-03,
   3.2660687009222814441e-01, 2.8982998459149642556e-03,
   3.2950362982175289764e-01, 2.8952150733304507490e-03,
   3.3239729116412812458e-01, 2.8921030764717064173e-03,
   3.3528782690968963080e-01, 2.8889638846014470666e-03,
   3.3817520987816382073e-01, 2.8857975272381085212e-03,
   3.4105941291892327906e-01, 2.8826040341555690561e-03,
   3.4394040891124207345e-01, 2.8793834353828694270e-03,
   3.4681817076455077597e-01, 2.8761357612039305018e-03,
   3.4969267141869120111e-01, 2.8728610421572684948e-03,
   3.5256388384417085764e-01, 2.8695593090357078067e-03,
   3.5543178104241711232e-01, 2.8662305928860914743e-03,
   3.5829633604603106270e-01, 2.8628749250089892305e-03,
   3.6115752191904111689e-01, 2.8594923369584031789e-03,
   3.6401531175715627774e-01, 2.8560828605414710857e-03,
   3.6686967868801912921e-01, 2.8526465278181672904e-03,
   3.6972059587145852233e-01, 2.8491833711010012403e-03,
   3.7256803649974195867e-01, 2.8456934229547136489e-03,
   3.7541197379782766863e-01, 2.8421767161959702838e-03,
   3.7825238102361638244e-01, 2.8386332838930533849e-03,
   3.8108923146820279134e-01, 2.8350631593655507170e-03,
   3.8392249845612669665e-01, 2.8314663761840422592e-03,
   3.8675215534562384434e-01, 2.8278429681697845341e-03,
   3.8957817552887644277e-01, 2.8241929693943925797e-03,
   3.9240053243226336119e-01, 2.8205164141795195677e-03,
   3.9521919951661000673e-01, 2.8168133370965340703e-03,
   3.9803415027743787743e-01, 2.8130837729661949783e-03,
   4.0084535824521378905e-01, 2.8093277568583240753e-03,
   4.0365279698559877327e-01, 2.8055453240914762690e-03,
   4.0645644009969664496e-01, 2.8017365102326074840e-03,
   4.0925626122430223619e-01, 2.7979013510967402185e-03,
   4.1205223403214929455e-01, 2.7940398827466267693e-03,
   4.1484433223215804366e-01, 2.7901521414924101257e-03,
   4.1763252956968240331e-01, 2.7862381638912825391e-03,
   4.2041679982675686702e-01, 2.7822979867471417677e-03,
   4.2319711682234303472e-01, 2.7783316471102450030e-03,
   4.2597345441257579821e-01, 2.7743391822768604783e-03,
   4.2874578649100917698e-01, 2.7703206297889167653e-03,
   4.3151408698886180228e-01, 2.7662760274336497593e-03,
   4.3427832987526204698e-01, 2.7622054132432473587e-03,
   4.3703848915749279891e-01, 2.7581088254944918412e-03,
   4.3979453888123587550e-01, 2.7539863027083999393e-03,
   4.4254645313081607734e-01, 2.7498378836498606196e-03,
   4.4529420602944487827e-01, 2.7456636073272705694e-03,
   4.4803777173946374996e-01, 2.7414635129921673928e-03,
   4.5077712446258711848e-01, 2.7372376401388605207e-03,
   4.5351223844014495055e-01, 2.7329860285040598383e-03,
   4.5624308795332496743e-01, 2.7287087180665020332e-03,
   4.5896964732341448395e-01, 2.7244057490465746668e-03,
   4.6169189091204187041e-01, 2.7200771619059379750e-03,
   4.6440979312141763527e-01, 2.7157229973471443987e-03,
   4.6712332839457512616e-01, 2.7113432963132558500e-03,
   4.6983247121561084703e-01, 2.7069380999874587164e-03,
   4.7253719610992438918e-01, 2.7025074497926766074e-03,
   4.7523747764445797396e-01, 2.6980513873911808464e-03,
   4.7793329042793560473e-01, 2.6935699546841987126e-03,
   4.8062460911110182605e-01, 2.6890631938115194352e-03,
   4.8331140838696008766e-01, 2.6845311471510979447e-03,
   4.8599366299101071117e-01, 2.6799738573186563850e-03,
   4.8867134770148845702e-01, 2.6753913671672833892e-03,
   4.9134443733959968976e-01, 2.6707837197870311237e-03,
   4.9401290676975913912e-01, 2.6661509585045101038e-03,
   4.9667673089982625485e-01, 2.6614931268824817855e-03,
   4.9933588468134115307e-01, 2.6568102687194489358e-03,
   5.0199034310976015178e-01, 2.6521024280492437873e-03,
   5.0464008122469089354e-01, 2.6473696491406139791e-03,
   5.0728507411012705288e-01, 2.6426119764968062895e-03,
   5.0992529689468262642e-01, 2.6378294548551481626e-03,
   5.1256072475182580335e-01, 2.6330221291866270352e-03,
   5.1519133290011241420e-01, 2.6281900446954674652e-03,
   5.1781709660341895561e-01, 2.6233332468187060677e-03,
   5.2043799117117518892e-01, 2.6184517812257642619e-03,
   5.2305399195859631044e-01, 2.6135456938180188319e-03,
   5.2566507436691469122e-01, 2.6086150307283703078e-03,
   5.2827121384361118403e-01, 2.6036598383208091685e-03,
   5.3087238588264599554e-01, 2.5986801631899798721e-03,
   5.3346856602468912142e-01, 2.5936760521607427178e-03,
   5.3605972985735034216e-01, 2.5886475522877335418e-03,
   5.3864585301540877759e-01, 2.5835947108549212540e-03,
   5.4122691118104199784e-01, 2.5785175753751632173e-03,
   5.4380288008405468854e-01, 2.5734161935897584747e-03,
   5.4637373550210686824e-01, 2.5682906134679988291e-03,
   5.4893945326094165585e-01, 2.5631408832067177781e-03,
   5.5150000923461258584e-01, 2.5579670512298373099e-03,
   5.5405537934571046931e-01, 2.5527691661879125638e-03,
   5.5660553956558979853e-01, 2.5475472769576743595e-03,
   5.5915046591459469302e-01, 2.5423014326415695994e-03,
   5.6169013446228438495e-01, 2.5370316825672995490e-03,
   5.6422452132765824178e-01, 2.5317380762873559984e-03,
   5.6675360267938032394e-01, 2.5264206635785553113e-03,
   5.6927735473600347558e-01, 2.5210794944415703629e-03,
   5.7179575376619294616e-01, 2.5157146191004603746e-03,
   5.7430877608894954086e-01, 2.5103260880021986467e-03,
   5.7681639807383229762e-01, 2.5049139518161981961e-03,
   5.7931859614118068883e-01, 2.4994782614338353016e-03,
   5.8181534676233634547e-01, 2.4940190679679709626e-03,
   5.8430662645986430173e-01, 2.4885364227524702746e-03,
   5.8679241180777375788e-01, 2.4830303773417197268e-03,
   5.8927267943173835949e-01, 2.4775009835101424263e-03,
   5.9174740600931599076e-01, 2.4719482932517112532e-03,
   5.9421656827016808006e-01, 2.4663723587794599504e-03,
   5.9668014299627841542e-01, 2.4607732325249921552e-03,
   5.9913810702217146813e-01, 2.4551509671379883738e-03,
   6.0159043723513022222e-01, 2.4495056154857109065e-03,
   6.0403711057541350786e-01, 2.4438372306525067265e-03,
   6.0647810403647283665e-01, 2.4381458659393083173e-03,
   6.0891339466516873667e-01, 2.4324315748631324732e-03,
   6.1134295956198658535e-01, 2.4266944111565770692e-03,
   6.1376677588125193809e-01, 2.4209344287673158020e-03,
   6.1618482083134535064e-01, 2.4151516818575909100e-03,
   6.1859707167491669310e-01, 2.4093462248037038748e-03,
   6.2100350572909895376e-01, 2.4035181121955041103e-03,
   6.2340410036572153043e-01, 2.3976673988358756440e-03,
   6.2579883301152300767e-01, 2.3917941397402217941e-03,
   6.2818768114836341751e-01, 2.3858983901359478493e-03,
   6.3057062231343598197e-01, 2.3799802054619417548e-03,
   6.3294763409947833520e-01, 2.3740396413680528093e-03,
   6.3531869415498322339e-01, 2.3680767537145683787e-03,
   6.3768378018440868034e-01, 2.3620915985716886307e-03,
   6.4004286994838767683e-01, 2.3560842322189992961e-03,
   6.4239594126393724171e-01, 2.3500547111449424607e-03,
   6.4474297200466705287e-01, 2.3440030920462853930e-03,
   6.4708394010098749600e-01, 2.3379294318275874141e-03,
   6.4941882354031718926e-01, 2.3318337876006648124e-03,
   6.5174760036728997192e-01, 2.3257162166840538103e-03,
   6.5407024868396135492e-01, 2.3195767766024715869e-03,
   6.5638674665001443157e-01, 2.3134155250862753614e-03,
   6.5869707248296524630e-01, 2.3072325200709195436e-03,
   6.6100120445836761966e-01, 2.3010278196964109553e-03,
   6.6329912091001742750e-01, 2.2948014823067621287e-03,
   6.6559080023015633253e-01, 2.2885535664494426858e-03,
   6.6787622086967496634e-01, 2.2822841308748288054e-03,
   6.7015536133831555987e-01, 2.2759932345356507817e-03,
   6.7242820020487402051e-01, 2.2696809365864386804e-03,
   6.7469471609740145390e-01, 2.2633472963829660968e-03,
   6.7695488770340512858e-01, 2.2569923734816920218e-03,
   6.7920869377004888153e-01, 2.2506162276392008215e-03,
   6.8145611310435296269e-01, 2.2442189188116403333e-03,
   6.8369712457339331678e-01, 2.2378005071541580876e-03,
   6.8593170710450030028e-01, 2.2313610530203356562e-03,
   6.8815983968545683187e-01, 2.2249006169616211364e-03,
   6.9038150136469597443e-01, 2.2184192597267597736e-03,
   6.9259667125149794671e-01, 2.2119170422612227293e-03,
   6.9480532851618656290e-01, 2.2053940257066339981e-03,
   6.9700745239032509810e-01, 2.1988502714001954821e-03,
   6.9920302216691157803e-01, 2.1922858408741102243e-03,
   7.0139201720057349102e-01, 2.1857007958550038097e-03,
   7.0357441690776192050e-01, 2.1790951982633439378e-03,
   7.0575020076694509609e-01, 2.1724691102128581720e-03,
   7.0791934831880136166e-01, 2.1658225940099498722e-03,
   7.1008183916641155828e-01, 2.1591557121531123157e-03,
   7.1223765297545082045e-01, 2.1524685273323410114e-03,
   7.1438676947437978378e-01, 2.1457611024285442135e-03,
   7.1652916845463520219e-01, 2.1390335005129516400e-03,
   7.1866482977081997302e-01, 2.1322857848465214018e-03,
   7.2079373334089256814e-01, 2.1255180188793451473e-03,
   7.2291585914635586929e-01, 2.1187302662500514289e-03,
   7.2503118723244540598e-01, 2.1119225907852072963e-03,
   7.2713969770831699402e-01, 2.1050950564987181231e-03,
   7.2924137074723377299e-01, 2.0982477275912256714e-03,
   7.3133618658675264100e-01, 2.0913806684495044003e-03,
   7.3342412552891008476e-01, 2.0844939436458560250e-03,
   7.3550516794040740338e-01, 2.0775876179375023304e-03,
   7.3757929425279532417e-01, 2.0706617562659762465e-03,
   7.3964648496265800856e-01, 2.0637164237565111901e-03,
   7.4170672063179644657e-01, 2.0567516857174286800e-03,
   7.4375998188741123796e-01, 2.0497676076395242297e-03,
   7.4580624942228475849e-01, 2.0427642551954515247e-03,
   7.4784550399496270946e-01, 2.0357416942391048896e-03,
   7.4987772642993504886e-01, 2.0286999908050000513e-03,
   7.5190289761781630247e-01, 2.0216392111076532034e-03,
   7.5392099851552525313e-01, 2.0145594215409583780e-03,
   7.5593201014646400656e-01, 2.0074606886775631311e-03,
   7.5793591360069643205e-01, 2.0003430792682425467e-03,
   7.5993269003512597629e-01, 1.9932066602412715667e-03,
   7.6192232067367284865e-01, 1.9860514987017956508e-03,
   7.6390478680745057641e-01, 1.9788776619311997736e-03,
   7.6588006979494192802e-01, 1.9716852173864757651e-03,
   7.6784815106217420295e-01, 1.9644742326995879989e-03,
   7.6980901210289388642e-01, 1.9572447756768374356e-03,
   7.7176263447874066732e-01, 1.9499969142982240274e-03,
   7.7370899981942081767e-01, 1.9427307167168074884e-03,
   7.7564808982287993216e-01, 1.9354462512580664379e-03,
   7.7757988625547502592e-01, 1.9281435864192559231e-03,
   7.7950437095214598901e-01, 1.9208227908687633255e-03,
   7.8142152581658639611e-01, 1.9134839334454626590e-03,
   7.8333133282141366953e-01, 1.9061270831580672643e-03,
   7.8523377400833859431e-01, 1.8987523091844809062e-03,
   7.8712883148833418349e-01, 1.8913596808711472809e-03,
   7.8901648744180389214e-01, 1.8839492677323979371e-03,
   7.9089672411874917850e-01, 1.8765211394497986196e-03,
   7.9276952383893641071e-01, 1.8690753658714940398e-03,
   7.9463486899206311752e-01, 1.8616120170115510799e-03,
   7.9649274203792358138e-01, 1.8541311630493004368e-03,
   7.9834312550657377245e-01, 1.8466328743286767123e-03,
   8.0018600199849562190e-01, 1.8391172213575569553e-03,
   8.0202135418476063301e-01, 1.8315842748070976623e-03,
   8.0384916480719282842e-01, 1.8240341055110702429e-03,
   8.0566941667853103219e-01, 1.8164667844651949558e-03,
   8.0748209268259048497e-01, 1.8088823828264733222e-03,
   8.0928717577442379082e-01, 1.8012809719125190226e-03,
   8.1108464898048119420e-01, 1.7936626232008872833e-03,
   8.1287449539877018561e-01, 1.7860274083284027593e-03,
   8.1465669819901443427e-01, 1.7783753990904859184e-03,
   8.1643124062281204657e-01, 1.7707066674404779358e-03,
   8.1819810598379314857e-01, 1.7630212854889641021e-03,
   8.1995727766777679120e-01, 1.7553193255030957536e-03,
   8.2170873913292717668e-01, 1.7476008599059107300e-03,
   8.2345247390990920462e-01, 1.7398659612756523665e-03,
   8.2518846560204333643e-01, 1.7321147023450870267e-03,
   8.2691669788545977646e-01, 1.7243471560008201813e-03,
   8.2863715450925196861e-01, 1.7165633952826110423e-03,
   8.3034981929562940673e-01, 1.7087634933826857546e-03,
   8.3205467614006975758e-01, 1.7009475236450491562e-03,
   8.3375170901147029481e-01, 1.6931155595647951097e-03,
   8.3544090195229864252e-01, 1.6852676747874154134e-03,
   8.3712223907874282714e-01, 1.6774039431081072990e-03,
   8.3879570458086063594e-01, 1.6695244384710795200e-03,
   8.4046128272272828106e-01, 1.6616292349688570408e-03,
   8.4211895784258836748e-01, 1.6537184068415843296e-03,
   8.4376871435299716358e-01, 1.6457920284763272638e-03,
   8.4541053674097117293e-01, 1.6378501744063736542e-03,
   8.4704440956813300590e-01, 1.6298929193105323939e-03,
   8.4867031747085654980e-01, 1.6219203380124312385e-03,
   8.5028824516041143598e-01, 1.6139325054798132253e-03,
   8.5189817742310680282e-01, 1.6059294968238317367e-03,
   8.5350009912043435304e-01, 1.5979113872983442155e-03,
   8.5509399518921070401e-01, 1.5898782522992045381e-03,
   8.5667985064171902987e-01, 1.5818301673635540528e-03,
   8.5825765056584999395e-01, 1.5737672081691112886e-03,
   8.5982738012524197025e-01, 1.5656894505334603439e-03,
   8.6138902455942055262e-01, 1.5575969704133379580e-03,
   8.6294256918393735047e-01, 1.5494898439039192755e-03,
   8.6448799939050806945e-01, 1.5413681472381023085e-03,
   8.6602530064714987603e-01, 1.5332319567857911038e-03,
   8.6755445849831804458e-01, 1.5250813490531776216e-03,
   8.6907545856504188570e-01, 1.5169164006820223330e-03,
   8.7058828654505995446e-01, 1.5087371884489335425e-03,
   8.7209292821295453743e-01, 1.5005437892646454426e-03,
   8.7358936942028541700e-01, 1.4923362801732949073e-03,
   8.7507759609572291199e-01, 1.4841147383516970308e-03,
   8.7655759424518019308e-01, 1.4758792411086194190e-03,
   8.7802934995194487200e-01, 1.4676298658840552400e-03,
   8.7949284937680986302e-01, 1.4593666902484950408e-03,
   8.8094807875820351583e-01, 1.4510897919021973371e-03,
   8.8239502441231901819e-01, 1.4427992486744579821e-03,
   8.8383367273324306755e-01, 1.4344951385228783230e-03,
   8.8526401019308381002e-01, 1.4261775395326321501e-03,
   8.8668602334209804586e-01, 1.4178465299157314470e-03,
   8.8809969880881770002e-01, 1.4095021880102909474e-03,
   8.8950502330017555668e-01, 1.4011445922797915074e-03,
   8.9090198360163025657e-01, 1.3927738213123422970e-03,
   8.9229056657729055586e-01, 1.3843899538199418219e-03,
   8.9367075917003884560e-01, 1.3759930686377377784e-03,
   8.9504254840165393025e-01, 1.3675832447232857518e-03,
   8.9640592137293306454e-01, 1.3591605611558067630e-03,
   8.9776086526381324711e-01, 1.3507250971354436709e-03,
   8.9910736733349177015e-01, 1.3422769319825164387e-03,
   9.0044541492054602362e-01, 1.3338161451367762690e-03,
   9.0177499544305255312e-01, 1.3253428161566586166e-03,
   9.0309609639870537015e-01, 1.3168570247185350853e-03,
   9.0440870536493351377e-01, 1.3083588506159642152e-03,
   9.0571280999901786246e-01, 1.2998483737589411688e-03,
   9.0700839803820719514e-01, 1.2913256741731463215e-03,
   9.0829545729983350021e-01, 1.2827908319991927651e-03,
   9.0957397568142653157e-01, 1.2742439274918727295e-03,
   9.1084394116082761054e-01, 1.2656850410194029319e-03,
   9.1210534179630267255e-01, 1.2571142530626688591e-03,
   9.1335816572665455761e-01, 1.2485316442144679896e-03,
   9.1460240117133454352e-01, 1.2399372951787519645e-03,
   9.1583803643055312063e-01, 1.2313312867698677126e-03,
   9.1706505988539000726e-01, 1.2227136999117975375e-03,
   9.1828345999790340472e-01, 1.2140846156373981740e-03,
   9.1949322531123849084e-01, 1.2054441150876388206e-03,
   9.2069434444973515097e-01, 1.1967922795108381552e-03,
   9.2188680611903494565e-01, 1.1881291902619003419e-03,
   9.2307059910618731355e-01, 1.1794549288015500354e-03,
   9.2424571227975500918e-01, 1.1707695766955663899e-03,
   9.2541213458991877389e-01, 1.1620732156140160808e-03,
   9.2656985506858123953e-01, 1.1533659273304853456e-03,
   9.2771886282947006361e-01, 1.1446477937213110513e-03,
   9.2885914706824029509e-01, 1.1359188967648107958e-03,
   9.2999069706257596973e-01, 1.1271793185405120502e-03,
   9.3111350217229093414e-01, 1.1184291412283803494e-03,
   9.3222755183942889759e-01, 1.1096684471080465391e-03,
   9.3333283558836271048e-01, 1.1008973185580330843e-03,
   9.3442934302589286879e-01, 1.0921158380549794491e-03,
   9.3551706384134524335e-01, 1.0833240881728665534e-03,
   9.3659598780666803314e-01, 1.0745221515822403145e-03,
   9.3766610477652794172e-01, 1.0657101110494342805e-03,
   9.3872740468840557574e-01, 1.0568880494357913638e-03,
   9.3977987756269006486e-01, 1.0480560496968846801e-03,
   9.4082351350277290194e-01, 1.0392141948817375023e-03,
   9.4185830269514100289e-01, 1.0303625681320423357e-03,
   9.4288423540946898499e-01, 1.0215012526813791214e-03,
   9.4390130199871066312e-01, 1.0126303318544325763e-03,
   9.4490949289918976284e-01, 1.0037498890662086759e-03,
   9.4590879863068984951e-01, 9.9486000782125028888e-04,
   9.4689920979654347271e-01, 9.8596077171285196884e-04,
   9.4788071708372052488e-01, 9.7705226442227391223e-04,
   9.4885331126291581371e-01, 9.6813456971795508907e-04,
   9.4981698318863584702e-01, 9.5920777145472555417e-04,
   9.5077172379928482975e-01, 9.5027195357301794603e-04,
   9.5171752411724987193e-01, 9.4132720009807818111e-04,
   9.5265437524898540695e-01, 9.3237359513917535076e-04,
   9.5358226838509681939e-01, 9.2341122288881082823e-04,
   9.5450119480042328150e-01, 9.1444016762192659336e-04,
   9.5541114585411979767e-01, 9.0546051369511278225e-04,
   9.5631211298973845600e-01, 8.9647234554581446953e-04,
   9.5720408773530888638e-01, 8.8747574769153769062e-04,
   9.5808706170341792408e-01, 8.7847080472905471155e-04,
   9.5896102659128847836e-01, 8.6945760133360855371e-04,
   9.5982597418085760512e-01, 8.6043622225811678130e-04,
   9.6068189633885378310e-01, 8.5140675233237455870e-04,
   9.6152878501687339266e-01, 8.4236927646225698553e-04,
   9.6236663225145639659e-01, 8.3332387962892071692e-04,
   9.6319543016416122221e-01, 8.2427064688800487638e-04,
   9.6401517096163884395e-01, 8.1520966336883126913e-04,
   9.6482584693570606592e-01, 8.0614101427360390321e-04,
   9.6562745046341800353e-01, 7.9706478487660782615e-04,
   9.6641997400713976368e-01, 7.8798106052340728480e-04,
   9.6720341011461732277e-01, 7.7888992663004321586e-04,
   9.6797775141904760187e-01, 7.6979146868223007491e-04,
   9.6874299063914773834e-01, 7.6068577223455201150e-04,
   9.6949912057922355337e-01, 7.5157292290965839807e-04,
   9.7024613412923721473e-01, 7.4245300639745872041e-04,
   9.7098402426487409399e-01, 7.3332610845431683739e-04,
   9.7171278404760881783e-01, 7.2419231490224461780e-04,
   9.7243240662477051260e-01, 7.1505171162809496199e-04,
   9.7314288522960724156e-01, 7.0590438458275421632e-04,
   9.7384421318134963435e-01, 6.9675041978033398824e-04,
   9.7453638388527370788e-01, 6.8758990329736236982e-04,
   9.7521939083276287817e-01, 6.7842292127197457802e-04,
   9.7589322760136916259e-01, 6.6924955990310301938e-04,
   9.7655788785487357181e-01, 6.6006990544966678759e-04,
   9.7721336534334569103e-01, 6.5088404422976060186e-04,
   9.7785965390320244981e-01, 6.4169206261984319461e-04,
   9.7849674745726608010e-01, 6.3249404705392515670e-04,
   9.7912464001482126177e-01, 6.2329008402275624882e-04,
   9.7974332567167145519e-01, 6.1408026007301218765e-04,
   9.8035279861019442043e-01, 6.0486466180648091561e-04,
   9.8095305309939692234e-01, 5.9564337587924836320e-04,
   9.8154408349496862125e-01, 5.8641648900088371326e-04,
   9.8212588423933514866e-01, 5.7718408793362417649e-04,
   9.8269844986171036742e-01, 5.6794625949155928814e-04,
   9.8326177497814781602e-01, 5.5870309053981473607e-04,
   9.8381585429159133649e-01, 5.4945466799373573071e-04,
   9.8436068259192488539e-01, 5.4020107881806992820e-04,
   9.8489625475602152753e-01, 5.3094241002614991828e-04,
   9.8542256574779161203e-01, 5.2167874867907528965e-04,
   9.8593961061823013010e-01, 5.1241018188489428605e-04,
   9.8644738450546325441e-01, 5.0313679679778506774e-04,
   9.8694588263479405947e-01, 4.9385868061723659397e-04,
   9.8743510031874742270e-01, 4.8457592058722914411e-04,
   9.8791503295711410590e-01, 4.7528860399541449668e-04,
   9.8838567603699401666e-01, 4.6599681817229578804e-04,
   9.8884702513283864958e-01, 4.5670065049040707557e-04,
   9.8929907590649270680e-01, 4.4740018836349263361e-04,
   9.8974182410723489781e-01, 4.3809551924568601507e-04,
   9.9017526557181791815e-01, 4.2878673063068891714e-04,
   9.9059939622450760694e-01, 4.1947391005094989670e-04,
   9.9101421207712128305e-01, 4.1015714507684298965e-04,
   9.9141970922906525985e-01, 4.0083652331584629973e-04,
   9.9181588386737153864e-01, 3.9151213241172063637e-04,
   9.9220273226673368067e-01, 3.8218406004368829931e-04,
   9.9258025078954185818e-01, 3.7285239392561213088e-04,
   9.9294843588591708461e-01, 3.6351722180517498655e-04,
   9.9330728409374462458e-01, 3.5417863146305981352e-04,
   9.9365679203870658441e-01, 3.4483671071213057761e-04,
   9.9399695643431368400e-01, 3.3549154739661434563e-04,
   9.9432777408193621167e-01, 3.2614322939128491892e-04,
   9.9464924187083416351e-01, 3.1679184460064853179e-04,
   9.9496135677818656976e-01, 3.0743748095813228770e-04,
   9.9526411586912001138e-01, 2.9808022642527622175e-04,
   9.9555751629673633096e-01, 2.8872016899093017276e-04,
   9.9584155530213954355e-01, 2.7935739667045705673e-04,
   9.9611623021446195481e-01, 2.6999199750494470128e-04,
   9.9638153845088949652e-01, 2.6062405956042920328e-04,
   9.9663747751668629280e-01, 2.5125367092713391391e-04,
   9.9688404500521847549e-01, 2.4188091971872980444e-04,
   9.9712123859797727384e-01, 2.3250589407162537390e-04,
   9.9734905606460141355e-01, 2.2312868214429782683e-04,
   9.9756749526289887452e-01, 2.1374937211668260962e-04,
   9.9777655413886807733e-01, 2.0436805218964657904e-04,
   9.9797623072671859987e-01, 1.9498481058458278992e-04,
   9.9816652314889157271e-01, 1.8559973554318500629e-04,
   9.9834742961607997465e-01, 1.7621291532749252492e-04,
   9.9851894842724916543e-01, 1.6682443822034952800e-04,
   9.9868107796965817762e-01, 1.5743439252651389306e-04,
   9.9883381671888259644e-01, 1.4804286657480799765e-04,
   9.9897716323884037566e-01, 1.3864994872198617512e-04,
   9.9911111618182284623e-01, 1.2925572735951552663e-04,
   9.9923567428853481652e-01, 1.1986029092546958274e-04,
   9.9935083638815074867e-01, 1.1046372792574375656e-04,
   9.9945660139840004927e-01, 1.0106612697302760146e-04,
   9.9955296832570700650e-01, 9.1667576861366910725e-05,
   9.9963993626543824646e-01, 8.2268166716457275281e-05,
   9.9971750440237472843e-01, 7.2867986319027466137e-05,
   9.9978567201168896283e-01, 6.3467126859804422993e-05,
   9.9984443846117119161e-01, 5.4065682893940007199e-05,
   9.9989380321694198787e-01, 4.4663758128575339384e-05,
   9.9993376586061777112e-01, 3.5261485987198697507e-05,
   9.9996432615388945509e-01, 2.5859124676461858672e-05,
   9.9998548438502844477e-01, 1.6457727579896868107e-05,
   9.9999724505455844035e-01, 7.0700764101825898713e-06};

} // end namespace Spheral
